<?php

/**
 * Define cron schedules functionality.
 *
 * Plugin cron schedules.
 *
 * @package    customily-options
 * @subpackage customily-options/cron
 * @author     Customily
 */

function wc_customily_init_cron_schedules()
{
    add_filter('cron_schedules', 'customily_20_minutes');
    add_action('wc_customily_order_creation', 'wc_customily_order_creation');
    require_once plugin_dir_path(dirname(__FILE__)) . '../includes/customily-options-utils.php';

    if (!wp_next_scheduled('wc_customily_order_creation')) {
        wp_schedule_event(time(), 'customily_20_minutes', 'wc_customily_order_creation');
        Customily_Options_Utils::add_log(
            "Customily Wordpress: initializing cron schedule task 'wc_customily_order_creation'."
        );
    }

    Customily_Options_Utils::create_table_orders_notifications();
}

function customily_20_minutes($schedules)
{
    $schedules['customily_20_minutes'] = array(
        'interval' => 20 * 60,
        'display' => __('Every 20 minutes', 'customily-v2')
    );
    return $schedules;
}

function wc_customily_order_creation()
{
    if (get_option('cl_cron_enabled') === 'y') {
        require_once plugin_dir_path(dirname(__FILE__)) . '../includes/customily-options-utils.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'db/customily-options-orders-notifications-db.php';

        $db = new Customily_Options_Orders_Notifications_DB();
        $orders = $db->get_orders_notifications(CUSTOMILY_OPTIONS_ORDER_CREATION_NOTIFY_QTY, 0);

        if (count($orders) > 0) {
            require_once plugin_dir_path(dirname(__FILE__)) . '../public/customily-options-public.php';
            $plugin_public = new Customily_Options_Public();
            foreach ($orders as $order) {
                $notified = $plugin_public->wc_customily_notify_order_creation($order['order_id']);
                Customily_Options_Utils::add_log(
                    "Customily Wordpress: Order notify task executed for order: {$order['order_id']}," .
                    " result {$notified}"
                );
                if ($notified == 'y') {
                    $db->delete_order_notification($order['order_id']);
                    Customily_Options_Utils::add_log("Customily Wordpress: Removed order notification for order: {$order['order_id']}");
                }
            }
        }
    }
}