<?php
if (!empty($_SERVER['REQUEST_METHOD']) && $_SERVER['REQUEST_METHOD'] == 'POST') {
    if (!empty($_POST)) {
        foreach ($_POST as $key => $value) {
            if (substr_count($key, 'cl_') > 0) {
                update_option(sanitize_text_field($key), sanitize_text_field($value));
            }
        }
    }
}

$cron_enabled = get_option('cl_cron_enabled');
if (empty($cron_enabled)) {
    $cron_enabled = 'y';
}

$check_products_connected = get_option('cl_check_products_connected');
if (empty($check_products_connected)) {
    $check_products_connected = 'n';
}

$customily_products_category = get_option('cl_products_category');
if (empty($customily_products_category)) {
    $customily_products_category = 'Customily';
}

$api_requests_timeout = get_option('cl_api_requests_timeout');
if (empty($api_requests_timeout)) {
    $api_requests_timeout = 180;
}

$api_key_option_name = CUSTOMILY_OPTIONS_ENV == 'beta' ? 'cl_beta_wp_api_key' : 'cl_wp_api_key';
?>
<style>
    .customily-tab {
        width: 100%;
        text-align: center;
        padding-top: 80px;
    }

    .customily-tab a,
    .customily-tab a:hover,
    .customily-tab a:active,
    .customily-tab a:focus,
    .customily-tab a:visited {
        box-shadow: none;
        outline: none;
        text-decoration: none;
    }

    .customily-tab .btn:hover,
    .customily-tab .btn:focus {
        background-color: #5b44bd !important;
        border-color: #5b44bd !important;
    }

    .customily-tab .btn {
        color: #fff;
        background-color: #7258e0;
        display: inline-block;
        font-weight: 700;
        text-align: center;
        vertical-align: middle;
        cursor: pointer;
        -webkit-user-select: none;
        user-select: none;
        border: 1px solid transparent;
        margin-top: 10px;
        padding: .375rem .75rem;
        font-size: 1rem;
        line-height: 1.5;
        border-radius: .25rem;
        transition: color .15s ease-in-out,background-color .15s ease-in-out,border-color .15s ease-in-out,box-shadow .15s ease-in-out;
    }

    .customily-tab h2 {
        font-size: 23px;
    }

    .customily-tab p {
        font-size: 18px;
    }

    .customily-tab label small {
        padding-left: 24px;
    }

    #cl_launch_link,
    .error-details {
        display: none;
    }

    .customily-tab .settings {
        margin-top: 20px;
        text-align: left;
        padding-left: 30px;
    }

    .customily-tab .settings input[type="text"] {
        padding: 5px 10px;
    }
</style>
<div class="customily-tab">
    <a href="<?php echo esc_url(CUSTOMILY_OPTIONS_BACKEND_URL) ?>/store" target="_blank">
        <img src="<?php echo esc_url(CUSTOMILY_OPTIONS_ASSETS_ROOT_URL) ?>/static/images/customily-logo.png"/>
    </a>
    <?php
        $error_message = get_option('cl_registration_error_message');
        if (!empty($error_message)) {
            $error_details = get_option('cl_registration_error_details');
    ?>
            <p><?php echo esc_html($error_message); ?></p>
            <small class="error-details"><?php echo esc_html($error_details); ?></small>
    <?php
        } else {
    ?>
        <p>Welcome Customily Wordpress App, now you can manage everything from app.customily.com</p>
        <button id="cl_launch_btn" type="button" class="btn">Launch Customily</button>
        <a id="cl_launch_link" href="<?php echo esc_url(CUSTOMILY_OPTIONS_BACKEND_URL) ?>/store" target="_blank"></a>
    <?php
        }
        $page = CUSTOMILY_OPTIONS_ENV == 'beta' ? 'customily_beta' : 'customily';
    ?>
    <div class="settings">
        <h2>Customily Store Settings</h2>
        <form method="post" action="./admin.php?page=<?php echo esc_attr($page); ?>">
            <p>
                <input id="cl_cron_enabled" type="checkbox" <?php echo $cron_enabled == 'y' ? 'checked' : '' ?> />
                <label for="cl_cron_enabled">Enable customily cron</label>
                <input name="cl_cron_enabled" type="hidden" value="<?php echo esc_attr($cron_enabled) ?>" />
            </p>
            <p>
                <input id="cl_check_products_connected" type="checkbox" <?php echo $check_products_connected == 'y' ? 'checked' : '' ?> />
                <label for="cl_check_products_connected">
                    Load Customily only in products that are part of the category: <b><?php echo esc_html($customily_products_category) ?></b> <br/>
                    <small>(Make sure you have the category created, and products assigned to it before enabling this option)</small>
                </label>
                <input name="cl_check_products_connected" type="hidden" value="<?php echo esc_attr($check_products_connected) ?>" />
            </p>
            <p>
                <label for="cl_products_category">Customily products category</label>
                <input name="cl_products_category" type="text" value="<?php echo esc_attr($customily_products_category) ?>" />
            </p>
            <p>
                <label for="cl_api_requests_timeout">Api requests timeout in seconds</label>
                <input name="cl_api_requests_timeout" type="text" value="<?php echo esc_attr($api_requests_timeout) ?>" />
            </p>
            <p style="display: none;">
                <label for="<?php echo esc_attr($api_key_option_name) ?>">Api Key</label>
                <input name="<?php echo esc_attr($api_key_option_name) ?>" type="text" value="<?php echo esc_attr(get_option($api_key_option_name)); ?>" style="width: 800px" />
            </p>
            <button class="btn" type="submit">Save Settings</button>
        </form>
    </div>
</div>
<script>
    window.addEventListener('DOMContentLoaded', function() {
        jQuery(function($) {
            $('#cl_launch_btn').on('click', function () {
                $('#cl_launch_link').get(0).click();
            });

            $('#cl_cron_enabled').on('click', function () {
                $('[name="cl_cron_enabled"]').val($(this).prop('checked') ? 'y' : 'n');
            });

            $('#cl_check_products_connected').on('click', function () {
                $('[name="cl_check_products_connected"]').val($(this).prop('checked') ? 'y' : 'n');
            });
        });
    });
</script>