<?php

/**
 * Define rest api functionality.
 *
 * Plugin rest api.
 *
 * @since      1.0.0
 * @package    customily-options
 * @subpackage customily-options/api
 * @author     Customily
 */
class Customily_Options_Api_Server extends WP_REST_Controller
{
    /* @var Customily_Options_Products_DB $products_db */
    private $products_db;
    /* @var Customily_Options_Settings_DB $options_db */
    private $options_db;
    /* @var Customily_Options_Orders_DB $orders_db */
    private $orders_db;

    public function __construct()
    {
        set_time_limit(600);
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-products-db.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-settings-db.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-orders-db.php';
        require_once 'customily-options-utils.php';
        $this->products_db = new Customily_Options_Products_DB();
        $this->options_db = new Customily_Options_Settings_DB();
        $this->orders_db = new Customily_Options_Orders_DB();
    }

    public function register_routes()
    {
        $namespace = 'customily/v1';
        register_rest_route($namespace, '/status', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'status'),
                'permission_callback' => array($this, 'api_public_access')
            )
        ]);
        register_rest_route($namespace, '/status/full', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'status_full'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/store/owner', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_store_owner_info'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/options', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_options'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/options', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'update_options'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/posts/meta', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_post_meta'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/posts/meta', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'update_post_meta'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_products'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'create_product'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/(?P<id>[\d]+)', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'update_product'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/(?P<id>[\d]+)/connection', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'update_product_connection'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/(?P<id>[\d]+)/connection', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_product_connection'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/(?P<id>[\d]+)', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_product'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/count', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_products_count'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/config', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_products_config'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/config/count', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_products_config_count'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/config/old', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_products_old_config'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/previews', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'save_preview'),
                'permission_callback' => array($this, 'api_public_access')
            )
        ]);
        register_rest_route($namespace, '/products/sync', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'sync_products'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/products/(?P<id>[\d]+)/url', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_product_url'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/product/url', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_product_url'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/(?P<id>[\d]+)', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_order'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_orders_between_ids'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders_by_dates', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_orders_between_dates'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/personalization/(?P<id>[\d]+)', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_order_personalization'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/(?P<id>[\d]+)/item/meta_data', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'update_order_item_meta'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/notify/creation', [
            array(
                'methods' => WP_REST_Server::READABLE,
                'callback' => array($this, 'get_orders_to_notify'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/notify/creation', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'notify_orders_creation'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/previews/fix', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'fix_previews_files'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/previews/extension', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'change_previews_extension'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
        register_rest_route($namespace, '/orders/fulfillment', [
            array(
                'methods' => WP_REST_Server::EDITABLE,
                'callback' => array($this, 'create_order_fulfillment'),
                'permission_callback' => array($this, 'api_restricted_access')
            )
        ]);
    }

    public function status()
    {
        $result = [
            'status' => 'active',
            'version' => CUSTOMILY_OPTIONS_PLUGIN_VERSION
        ];
        return new WP_REST_Response($result, 200);
    }

    public function status_full()
    {
        $memory_limit = ini_get('memory_limit');
        $memory_um = 'MB';
        if (preg_match('/^(\d+)(.)$/', $memory_limit, $matches)) {
            if ($matches[2] == 'G') {
                $memory_limit = $matches[1];
                $memory_um = 'GB';
            } else if ($matches[2] == 'M') {
                $memory_limit = $matches[1]; // nnnM -> nnn MB
            } else if ($matches[2] == 'K') {
                $memory_limit = $matches[1] / 1024; // nnnK -> nnn KB
            }
        }

        $result = [
            'status' => 'active',
            'version' => CUSTOMILY_OPTIONS_PLUGIN_VERSION,
            'php_settings' => [
                'memory_limit' => "{$memory_limit}{$memory_um}",
                'max_input_vars' => ini_get('max_input_vars'),
                'max_execution_time' => ini_get('max_execution_time')
            ],
            'plugin_settings' => [
                'remove_www_from_url' => get_option('cl_remove_www_from_url'),
                'cron_enabled' => get_option('cl_cron_enabled'),
                'check_products_connected' => get_option('cl_check_products_connected'),
                'products_category' => get_option('cl_products_category'),
                'api_requests_timeout' => get_option('cl_api_requests_timeout'),
                'preview_link_html' => get_option('cl_preview_link_html'),
                'show_customization_url_as_text' => get_option('cl_show_customization_url_as_text')
            ]
        ];
        return new WP_REST_Response($result, 200);
    }

    public function get_store_owner_info()
    {
        $result = $this->options_db->get_store_owner_info();
        return new WP_REST_Response($result, 200);
    }

    public function get_options($request)
    {
        $result = false;
        $setting_key = $request['setting_key'];
        if (isset($setting_key)) {
            $result = get_option($setting_key);
        }
        return new WP_REST_Response($result, 200);
    }

    public function update_options($request)
    {
        $result = ['success' => false];
        $setting_key = $request['setting_key'];
        $setting_value = $request['setting_value'];
        if (isset($setting_key) && isset($setting_value)) {
            update_option($setting_key, $setting_value);
            $result['success'] = true;
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_post_meta($request)
    {
        $result = false;
        $post_id = $request['id'];
        $meta_key = $request['meta_key'];
        if (isset($post_id) && isset($meta_key)) {
            $result = get_post_meta($post_id, $meta_key);
        }
        return new WP_REST_Response($result, 200);
    }

    public function update_post_meta($request)
    {
        $result = ['success' => false];
        $post_id = $request['id'];
        $meta_key = $request['meta_key'];
        $meta_value = $request['meta_value'];
        if (isset($post_id) && isset($meta_key) && isset($meta_value)) {
            update_post_meta($post_id, $meta_key, $meta_value);
            $result['success'] = true;
        }
        return new WP_REST_Response($result, 200);
    }

    public function create_product($request)
    {
        $result = ['product' => false];
        $product_info = [
            'name' => $request['title'] ?? 'Customily Product',
            'description' => $request['body_html'] ?? '',
            'published' => $request['published'] ?? false,
            'images' => $request['images'] ?? [],
            'options' => $request['options'] ?? [],
            'variants' => $request['variants'] ?? [],
            'vendor' => $request['vendor'] ?? ''
        ];

        Customily_Options_Utils::add_log("Customily Wordpress: creating product '{$product_info['name']}'.");

        if (isset($request['product_type']) && $request['product_type'] == 'from-provider') {
            $response = $this->products_db->create_product($product_info);
        } else {
            $response = $this->products_db->create_demo_product($product_info);
        }

        if (!empty($response)) {
            $result['product'] = $this->products_db->build_product_item($response);
            Customily_Options_Utils::add_log("Customily Wordpress: '{$product_info['name']}' product was created.");
        } else {
            Customily_Options_Utils::add_log("Customily Wordpress: '{$product_info['name']}' product is already exists.");
        }

        return new WP_REST_Response($result, 200);
    }

    public function update_product($request)
    {
        $result = ['success' => '0'];
        $product_id = 0;

        if (isset($request['id'])) {
            $product_id = (int)$request['id'];
        } else if (isset($request['product_id'])) {
            $product_id = (int)$request['product_id'];
        }

        $product_info = [
            'id' => $product_id,
            'name' => $request['title'],
            'description' => $request['body_html'] ?? '',
            'published' => $request['published'] ?? false,
            'image' => $request['image'] ?? false,
            'variants' => $request['variants'] ?? [],
            'vendor' => $request['vendor'] ?? ''
        ];

        Customily_Options_Utils::add_log("Customily Wordpress: updating product '{$product_info['name']}'.");

        $response = $this->products_db->update_product($product_info);

        if (!empty($response)) {
            $result['success'] = '1';
            Customily_Options_Utils::add_log("Customily Wordpress: '{$product_info['name']}' product was updated.");
        } else {
            Customily_Options_Utils::add_log("Customily Wordpress: '{$product_info['name']}' product not found.");
        }

        return new WP_REST_Response($result, 200);
    }

    public function get_product_connection($request)
    {
        $result = ['connected' => 'n'];
        $product_id = 0;

        if (isset($request['id'])) {
            $product_id = (int)$request['id'];
        } else if (isset($request['product_id'])) {
            $product_id = (int)$request['product_id'];
        }

        if ($product_id > 0) {
            $r = get_post_meta($product_id, CUSTOMILY_OPTIONS_PRODUCT_CONNECTED_KEY);
            if (!empty($r)) {
                $result['connected'] = $r[0];
            }
        }
        return new WP_REST_Response($result, 200);
    }

    public function update_product_connection($request)
    {
        $result = ['success' => '0'];
        $product_id = 0;

        if (isset($request['id'])) {
            $product_id = (int)$request['id'];
        } else if (isset($request['product_id'])) {
            $product_id = (int)$request['product_id'];
        }

        if ($product_id > 0) {
            $value = $request['connected'] ?? 'n';
            update_post_meta($product_id, CUSTOMILY_OPTIONS_PRODUCT_CONNECTED_KEY, $value);
            Customily_Options_Utils::add_log(
                "Customily Wordpress: Updating product connection for id: '{$product_id}', value {$value}"
            );
            $result['success'] = '1';
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_product($request)
    {
        $result = [];
        $product_id = 0;

        if (isset($request['id'])) {
            $product_id = (int)$request['id'];
        } else if (isset($request['product_id'])) {
            $product_id = (int)$request['product_id'];
        }

        if ($product_id > 0) {
            $result['product'] = $this->products_db->build_product_item($product_id);
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_products($request)
    {
        // execution time may be very long, which again depends on indexing mode (on-save or on-schedule) configured
        // phpcs:ignore
        set_time_limit(7200);
        // phpcs:ignore
        ini_set('memory_limit', '6G');

        $result = ['total' => 0, 'products' => []];
        $search_string = $request['search_string'] ?? '';
        $ids = $request['ids'] ?? '';
        $limit = $request['limit'] ?? 100;
        $offset = $request['offset'] ?? 0;
        try {
            $result['total'] = $this->products_db->get_products_count($search_string, $ids);
            $result['products'] = $this->products_db->get_products($search_string, $ids, $limit, $offset);
        } catch (Exception $e) {
            Customily_Options_Utils::add_log('Customily Wordpress: Error getting product from api');
            Customily_Options_Utils::add_log($e->getTraceAsString());
            return new WP_REST_Response(['error' => $e->getTraceAsString()], 500);
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_products_count()
    {
        $search_string = $request['search_string'] ?? '';
        $ids = $request['ids'] ?? '';
        $result = $this->products_db->get_products_count($search_string, $ids);
        return new WP_REST_Response($result, 200);
    }

    public function get_products_config($request)
    {
        $limit = $request['limit'] ?? 10;
        $offset = $request['offset'] ?? 0;
        $ids = $request['ids'] ?? '';
        $only_post_meta = $request['only_post_meta'] ?? false;
        $result['configs'] = $this->products_db->get_products_config($limit, $offset, $ids, $only_post_meta);
        return new WP_REST_Response($result, 200);
    }

    public function get_products_config_count($request)
    {
        $ids = $request['ids'] ?? '';
        $result['configs'] = $this->products_db->get_products_config_count($ids);
        return new WP_REST_Response($result, 200);
    }

    public function get_products_old_config($request)
    {
        $ids = $request['ids'];
        $result['configs'] = $this->products_db->get_products_old_config($ids);
        return new WP_REST_Response($result, 200);
    }

    public function get_product_url($request)
    {
        $result = ['url' => ''];
        $product_id = 0;

        if (isset($request['id'])) {
            $product_id = (int)$request['id'];
        } else if (isset($request['productId'])) {
            $product_id = (int)$request['productId'];
        }

        if ($product_id > 0) {
            $permalink = get_permalink($product_id);
            if ($permalink) {
                $result['url'] = $permalink;
            }
        }
        return new WP_REST_Response($result, 200);
    }

    public function save_preview($request)
    {
        $result = array(
            'url' => false
        );
        $data = $request['image'];
        if (isset($data) && !empty($data)) {
            $response = Customily_Options_Utils::save_file($data);
            if ($response['saved']) {
                $result['url'] = $response['url'];
            }
        }

        return new WP_REST_Response($result, 200);
    }

    public function sync_products($request)
    {
        $search_string = $request['search_string'] ?? '';
        $limit = $request['limit'] ?? 100;
        $offset = $request['offset'] ?? 0;
        $ids = $request['ids'] ?? '';
        $result['products'] = $this->products_db->sync_products($search_string, $limit, $offset, $ids);
        return new WP_REST_Response($result, 200);

    }

    public function get_order($request)
    {
        $result = ['order' => false];
        $order_id = $request['id'];
        $include_all_keys = $request['include_all_keys'] ?? '0';
        $production_file_full_url = $request['production_file_full_url'] ?? '0';
        if (isset($order_id)) {
            $order = $this->orders_db->get_order_by_id($order_id, $include_all_keys, $production_file_full_url);
            if (!empty($order)) {
                $result['order'] = $order;
            }
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_orders_between_ids($request)
    {
        $result = array(
            'orders' => false,
            'count' => 0
        );
        $from = $request['from'];
        $to = $request['to'];
        if (isset($from) && isset($to)) {
            $include_all_keys = $request['include_all_keys'] ?? '0';
            $orders = $this->orders_db->get_orders_between_ids((int)$from, (int)$to, $include_all_keys);
            $result['orders'] = $orders;
            $result['count'] = count($orders);
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_orders_between_dates($request)
    {
        $result = array(
            'orders' => false,
            'count' => 0
        );
        $from_date = $request['from_date'];
        $to_date = $request['to_date'];
        if (isset($from_date) && isset($to_date)) {
            $include_all_keys = $request['include_all_keys'] ?? '0';
            $orders = $this->orders_db->get_orders_between_dates($from_date, $to_date, $include_all_keys);
            $result['orders'] = $orders;
            $result['count'] = count($orders);
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_order_personalization($request)
    {
        $result = false;
        $order_id = $request['id'];
        if (isset($order_id)) {
            $include_all_keys = $request['include_all_keys'] ?? false;
            $result = $this->orders_db->get_customily_info($order_id, $include_all_keys);
        }
        return new WP_REST_Response($result, 200);
    }

    public function update_order_item_meta($request)
    {
        $result = ['success' => false];
        $order_id = $request['id'];
        $cart_item_key = $request['cart_item_key'];
        $meta_key = $request['meta_key'];
        $meta_value = $request['meta_value'];
        if (isset($order_id) && isset($cart_item_key) && isset($meta_key) && isset($meta_value)) {
            $result['success'] = $this->orders_db->update_order_item_meta(
                $order_id,
                $cart_item_key,
                $meta_key,
                $meta_value
            );
        }
        return new WP_REST_Response($result, 200);
    }

    public function get_orders_to_notify($request)
    {
        $start_time = time();
        $limit = $request['limit'] ?? 100;
        $offset = $request['offset'] ?? 0;
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-orders-notifications-db.php';
        $db = new Customily_Options_Orders_Notifications_DB();
        $orders = $db->get_orders_notifications($limit, $offset);
        $end_time = time();
        $result = array(
            'start_time' => gmdate('Y-m-d h:i:s', $start_time),
            'end_time' => gmdate('Y-m-d h:i:s', $end_time),
            'orders' => $orders
        );
        return new WP_REST_Response($result, 200);
    }

    public function notify_orders_creation($request)
    {
        $result = array();
        $ids = $request['ids'];
        if (isset($ids)) {
            require_once plugin_dir_path(dirname(__FILE__)) . 'public/customily-options-public.php';
            $plugin_public = new Customily_Options_Public();
            $ids = explode(',', $ids);
            foreach ($ids as $id) {
                $notified = $plugin_public->wc_customily_notify_order_creation($id, true);
                if ($notified == 'y') {
                    array_push($result, $id);
                }
            }
        }
        return new WP_REST_Response($result, 200);
    }

    public function fix_previews_files($request) {
        $result = array();
        $update_time = $request['update_time'];
        if ($update_time) {
            $result = Customily_Options_Utils::fix_previews_files((int) $update_time);
        }
        return new WP_REST_Response($result, 200);
    }

    public function change_previews_extension($request) {
        $result = array();
        $update_time = $request['update_time'];
        if ($update_time) {
            $result = Customily_Options_Utils::change_previews_extension((int) $update_time);
        }
        return new WP_REST_Response($result, 200);
    }

    public function create_order_fulfillment($requet) {
        $result = array(
            'fulfillment' => array(
                'id' => -1
            )
        );
        $line_items_by_fulfillment_order = $requet['line_items_by_fulfillment_order'];
        $notify_customer = $requet['notify_customer'];
        $tracking_info = $requet['tracking_info'];
        if (isset($line_items_by_fulfillment_order) && isset($notify_customer) && isset($tracking_info)) {
            $result['fulfillment']['id'] = $this->orders_db->create_order_fulfillment(
                $line_items_by_fulfillment_order,
                $notify_customer,
                $tracking_info
            );
        }
        return new WP_REST_Response($result, 200);
    }

    /**
     * @return bool
     */
    public function api_public_access() {
        return true;
    }

    /**
     * @param WP_REST_Request $request
     * @return bool
     */
    public function api_restricted_access($request)
    {
        if (CUSTOMILY_OPTIONS_ENV == 'dev') {
            return true;
        }
        $api_key = $request->get_header('wp-api-key');
        $api_key = $api_key ?? false;
        $stored_api_key = CUSTOMILY_OPTIONS_ENV == 'beta'
                        ? get_option('cl_beta_wp_api_key')
                        : get_option('cl_wp_api_key');

        if (!empty($api_key) && !empty($stored_api_key) && $api_key == $stored_api_key) {
            return true;
        }
        return false;
    }
}
