<?php
/**
 * Fired during plugin deactivation.
 *
 * This class defines all code necessary to run during the plugin's deactivation.
 *
 * @since      1.0.0
 * @package    customily-options
 * @subpackage customily-options/includes
 * @author     Customily
 */
class Customily_Options_Deactivator {
    public function __construct() { }

    /**
     * Deactivate plugin on site network
     *
     * @param $network_wide
     */
    public function deactivate($network_wide) {
        global $wpdb;

        // Check if the plugin is being network-activated or not.
        if ($network_wide) {
            // Retrieve all site IDs from this network (WordPress >= 4.6 provides easy to use functions for that).
            if (function_exists('get_sites') && function_exists('get_current_network_id')) {
                $site_ids = get_sites(array('fields' => 'ids', 'network_id' => get_current_network_id()));
            } else {
                $site_ids = $wpdb->get_col("SELECT blog_id FROM $wpdb->blogs WHERE site_id = $wpdb->site_id");
            }

            // Install the plugin for all these sites.
            foreach ($site_ids as $site_id) {
                switch_to_blog($site_id);
                $this->deactivate_single_site(Customily_Options_Utils::get_site_url($site_id));
                restore_current_blog();
            }
        } else {
            $this->deactivate_single_site(CUSTOMILY_OPTIONS_SITE_URL);
        }
    }

    /**
     * Deactivate plugin on single site
     */
    private function deactivate_single_site($site_url) {
        if (get_option('cl_remove_www_from_url') == 'y') {
            $site_url = preg_replace('#^www.#', '', $site_url);
        }

        $this->customily_unschedule_events();
        require_once 'customily-options-utils.php';
        Customily_Options_Utils::add_log("Customily Wordpress: Plugin deactivated successfully for shop '{$site_url}'.");
    }

    private function customily_unschedule_events() {
        $timestamp = wp_next_scheduled('wc_customily_order_creation');
        wp_unschedule_event($timestamp, 'wc_customily_order_creation');
    }
}
