<?php
/**
 * The public-facing functionality of the plugin.
 *
 * Defines the plugin name, version, and two examples hooks for how to
 * enqueue the public-facing stylesheet and JavaScript.
 *
 * @package    customily-options
 * @subpackage customily-options/public
 * @author     Customily
 */

const PROPERTIES_SEARCH_ARRAY = array(
    'property___customily-template-id',
    'property___customily-woocommerce-product-id',
    'property[_customily-production-url]',
    'property___customily-production-url',
    'property[_customily-personalization-id]',
    'property___customily-personalization-id',
    'property___customily-eps-name',
    'property___customily-eps-placeholder',
    'property___customily-file-numbers',
    'property[_customily-customization-url]',
    'property___customily-customization-url',
    'property[_customily-added-date]',
    'property___customily-added-date',
    'property[_customily-thumb]',
    'property___customily-thumb',
    'property[_customily-thumb-id]',
    'property___customily-thumb-id',
    'property___customily-thumb-base64',
    'property___customily-session-id',
    'cl_template_id',
    'cl_input-added-date',
    'cl_personalization_id'
);

const CUSTOMILY_SPACE_REPLACEMENT = '_spc_';

class Customily_Options_Public
{
    /**
     * Plugin api client
     *
     * @var Customily_Options_Api_Client
     */
    private $api_client;

    /**
     * The customily app settings.
     *
     * @since    1.0.0
     * @access   private
     * @var      mixed.
     */
    private $app_settings;

    private $can_load_customily;
    private $assets_inserted;
    private $options_container_inserted;
    private $site_url;
    private $custom_strings;

    /**
     * Initialize the class and set its properties.
     *
     * @since    1.0.0
     */
    public function __construct()
    {
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-products-db.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-orders-db.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'includes/customily-options-api-client.php';
        require_once plugin_dir_path(dirname(__FILE__)) . 'includes/customily-options-utils.php';

        $this->site_url = Customily_Options_Utils::get_site_url();
        $this->api_client = new Customily_Options_Api_Client($this->site_url);
        $this->can_load_customily = null;
        $this->assets_inserted = false;
        $this->options_container_inserted = false;
        $this->custom_strings = false;
        add_filter('kses_allowed_protocols', function ($protocols) {
            $protocols[] = 'data';
            return $protocols;
        });
    }

    /**
     * Insert customily preview in product page
     */
    public function wc_customily_insert_assets()
    {
        if ($this->assets_inserted) {
            return;
        }

        global $post;

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($post->ID);
        if ($this->can_load_customily === 'n') {
            return;
        }

        echo '<script> window.customilyWordpress = {}; </script>';

        $this->app_settings = !empty($this->app_settings['settings'])
            ? $this->app_settings
            : $this->api_client->get_unified_settings($post->post_name);

        if (
            isset($this->app_settings['settings']) &&
            (
                $this->app_settings['settings']['enabled'] ||
                $this->use_app_version2() == '1'
            ) &&
            isset($this->app_settings['productConfig']) &&
            count($this->app_settings['productConfig']['assignedSets']) > 0
        ) {
            wp_register_script(
                'customily-settings-script',
                '',
                array(),
                esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION),
                true
            );
            wp_enqueue_script(
                'customily-settings-script',
                '',
                array(),
                esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION),
                true
            );

            $products_db = new Customily_Options_Products_DB();
            $product = $products_db->build_product_item($post->ID, true, true, false);

            try {
                add_filter('wc_stripe_hide_payment_request_on_product_page', '__return_true');
                if (
                    isset($this->app_settings['settings']['debugMode']) &&
                    $this->app_settings['settings']['debugMode']
                ) {
                    Customily_Options_Utils::add_log(
                        "Customily Wordpress: Hiding payment request buttons on product page for '{$product['title']}'"
                    );
                }
            } catch (Exception $ex) {
                if (
                    isset($this->app_settings['settings']['debugMode']) &&
                    $this->app_settings['settings']['debugMode']
                ) {
                    Customily_Options_Utils::add_log(
                        "Customily Wordpress: Error hiding payment request buttons on product page for '{$product['title']}'"
                    );
                }
                Customily_Options_Utils::add_log($ex->getTraceAsString());
            }

            if ($this->use_app_version2() == '1') {
                $this->app_settings['settings']['enabled'] = true;
            }

            $settings = get_option('cl_dont_init_unified_settings') != 'y' ? $this->app_settings : '[]';

            wp_add_inline_script(
                'customily-settings-script',
                '
                    var productJson = ' . wp_json_encode($product) . ';
                    var unifiedSettings = ' . wp_json_encode($settings) . ';
                    window.customilyWordpress = {                       
                        appLoading: false,         
                        shop: "' . esc_html($this->site_url) . '",          
                        apiRoot: "' . esc_url(Customily_Options_Utils::get_api_root()) . '",
                        priceOptions: {
                            disabled: unifiedSettings.settings
                                    ? !unifiedSettings.settings.woocommerce?.optionsPrices?.enabled
                                    : false,
                            currencyPosition: "' . esc_html(get_option('woocommerce_currency_pos')) . '",
                            currencySymbol: "' . esc_html(get_woocommerce_currency_symbol()) . '",                            
                            thousandSeparator: "' . esc_html(wc_get_price_thousand_separator()) . '",                        
                            decimalSeparator: "' . esc_html(wc_get_price_decimal_separator()) . '",
                            decimalCount: "' . esc_html(wc_get_price_decimals()) . '",
                            originalContainer: null,
                            finalContainer: null,
                            modalOriginalContainer: null,
                            modalFinalContainer: null
                        },                        
                        cdn: "' . esc_url(CUSTOMILY_OPTIONS_ASSETS_ROOT_URL) . '",                        
                        product: productJson || {},
                        unifiedSettings: unifiedSettings,
                        livePreviewMode: undefined,
                        showPreviewByDefault: unifiedSettings && unifiedSettings.settings
                                            ? unifiedSettings.settings.showPreviewByDefault
                                            : false,
                        addTemplateIdToCartItem: true,
                        wpIsMobile: "' . esc_html(wp_is_mobile()) . '",
                        device: {},
                        sticky: {
                            enabled: unifiedSettings && unifiedSettings.settings
                                   ? unifiedSettings.settings.stickyPreview
                                   : false,
                            mobileEnabled: unifiedSettings && unifiedSettings.settings
                                   ? unifiedSettings.settings.stickyPreviewMobile
                                   : false,                                   
                            leftContainer: null,
                            rightContainer: null,
                            padding: 0
                        },
                        buttons: {
                            cartButtonSelector: undefined,
                            disabledUntilAppLoad: true,
                            disabledCount: 0,
                            disableMaximumTime: 10 * 1000
                        },
                        pluginVersion: "' . esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION) . '",
                        pluginEnv: "' . esc_html(CUSTOMILY_OPTIONS_ENV) . '"
                    };
                    customilyWordpress.product.connected = "y";
                    const isFirstInstallation = "' . esc_html(get_option('cl_first_installation')) . '";
                    window.customilyWordpress.dontChangeImagePositionOnHide = isFirstInstallation != "y";                
                '
            );

            if (isset($this->app_settings['settings']['customCss'])) {
                wp_register_style(
                    'customily-settings-style',
                    '',
                    array(),
                    esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION)
                );
                wp_enqueue_style(
                    'customily-settings-style',
                    '',
                    array(),
                    esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION)
                );
                wp_add_inline_style('customily-settings-style', $this->app_settings['settings']['customCss']);
            }

            if (isset($this->app_settings['settings']['customJs'])) {
                wp_add_inline_script('customily-settings-script', $this->app_settings['settings']['customJs']);
            }

            $this->app_initialize_script();

            wp_enqueue_script(
                'customily-js',
                esc_url(CUSTOMILY_OPTIONS_CDN_URL) . '/customily.js',
                array('customily-settings-script'),
                esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION),
                true
            );

            $script_url = esc_url(CUSTOMILY_OPTIONS_ASSETS_ROOT_URL) . '/static/scripts/';
            $script_url .= CUSTOMILY_OPTIONS_ENV == 'dev'
                ? 'customily.wordpress.script.unified.dev.js'
                : 'customily.wordpress.script.unified.js';

            wp_enqueue_script(
                'customily-wordpress-script',
                $script_url,
                array('customily-js'),
                esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION),
                true
            );

            if (
                isset($this->app_settings['settings']['debugMode']) &&
                $this->app_settings['settings']['debugMode']
            ) {
                Customily_Options_Utils::add_log("Customily Wordpress: Preview inserted in product page for '{$product['title']}'");
            }
        } else {
            echo '<script>
                    window.customilyWordpress = {};
                    window.customilyWordpress.shop = "' . esc_url($this->site_url) . '";              
                    window.customilyWordpress.apiRoot = "' . esc_url(Customily_Options_Utils::get_api_root()) . '";
                 </script>';
        }

        $this->assets_inserted = true;
    }

    /**
     *
     */
    public function wc_customily_single_product_summary_sibling()
    {
        echo '<div class="customily-product-summary-sibling hidden"></div>';
    }

    /**
     * Add class to gallery container
     * @param $classes
     * @return mixed
     */
    public function wc_customily_product_image_gallery_classes($classes)
    {
        $classes[] = 'customily-gallery-container';
        return $classes;
    }

    /**
     * Insert customily options container in product page
     */
    public function wc_customily_insert_options_container()
    {
        if ($this->options_container_inserted) {
            return;
        }

        global $post;

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($post->ID);
        if ($this->can_load_customily === 'n') {
            return;
        }

        // on wooper.ch action woocommerce_after_single_product isn't called
        if (!$this->assets_inserted) {
            $this->wc_customily_insert_assets();
        }

        $this->app_settings = isset($this->app_settings['settings'])
            ? $this->app_settings
            : $this->api_client->get_unified_settings($post->post_name);

        if (
            isset($this->app_settings['settings']) &&
            (
                $this->app_settings['settings']['enabled'] ||
                $this->use_app_version2() == '1'
            ) &&
            isset($this->app_settings['productConfig']) &&
            count($this->app_settings['productConfig']['assignedSets'])
        ) {
            echo '<div id="customily-options"></div>';
            if (
                isset($this->app_settings['settings']['debugMode']) &&
                $this->app_settings['settings']['debugMode']
            ) {
                Customily_Options_Utils::add_log(
                    "Customily Wordpress: Options container inserted in product page for '{$post->post_title}'"
                );
            }
        }
        $this->options_container_inserted = true;
    }

    /**
     * Add to cart button text for single products
     * @param $button_text
     * @param $product
     * @return mixed
     */
    public function wc_customily_handle_single_product_stock_status($button_text, $product) {
        if ($product->is_type('variable') || $product->is_in_stock()) {
            return $button_text;
        }

        global $post;

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($post->ID);
        if ($this->can_load_customily === 'n') {
            return $button_text;
        }

        $this->app_settings = isset($this->app_settings['settings'])
            ? $this->app_settings
            : $this->api_client->get_unified_settings($post->post_name);

        if (
            isset($this->app_settings['settings']) &&
            (
                $this->app_settings['settings']['enabled'] ||
                $this->use_app_version2() == '1'
            )
        ) {
            $handle_stock_status = $this->get_setting_value(
                'cl_handle_product_stock_status',
                $this->app_settings['settings']
            );

            if (!$handle_stock_status) {
                return $button_text;
            }

            $button_text = $this->get_setting_value('add_to_cart_text', $this->app_settings['settings']);
        }

        return $button_text;
    }

    /**
     * Handle out of stock for variable products
     */
    public function wc_customily_handle_variable_product_stock_status()
    {
        global $post, $product;

        if (!$product->is_type('variable')) {
            return;
        }

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($post->ID);
        if ($this->can_load_customily === 'n') {
            return;
        }

        $this->app_settings = isset($this->app_settings['settings'])
            ? $this->app_settings
            : $this->api_client->get_unified_settings($post->post_name);

        if (
            isset($this->app_settings['settings']) &&
            (
                $this->app_settings['settings']['enabled'] ||
                $this->use_app_version2() == '1'
            )
        ) {
            $handle_stock_status = $this->get_setting_value(
                'cl_handle_product_stock_status',
                $this->app_settings['settings']
            );

            if (!$handle_stock_status) {
                return;
            }

            $data = '[';

            // Loop through variation Ids
            foreach ($product->get_visible_children() as $variation_id) {
                $variation = wc_get_product($variation_id);
                $in_stock = $variation->is_in_stock() ? 'y' : 'n';
                $data .= "{ \"key\": \"{$variation_id}\", \"value\": \"{$in_stock}\"},";
            }

            $data = substr($data, 0, -1);
            $data .= ']';

            $change_button_text = $this->get_setting_value(
                'cl_change_button_text_by_product_stock_status',
                $this->app_settings['settings']
            );
            $out_of_stock_text = '';

            if ($change_button_text) {
                $out_of_stock_text = $this->get_setting_value('add_to_cart_text', $this->app_settings['settings']);
            }

            echo '
                <script type="text/javascript">
                    window.addEventListener("customily-options-loaded", function() {
                        jQuery(function($) {
                            var button = "button.single_add_to_cart_button",
                                buttonText = $(button).first().text(),
                                dataObj = {},
                                dataArray = "' . esc_html($data) . '";
                            dataArray = JSON.parse(dataArray);    
                            var changeButtonText = "' . esc_html($change_button_text) . '";
                    
                            if (dataArray) {
                                dataArray.forEach(function (o) {
                                    dataObj[o.key] = o.value;
                                });
                                dataObj["0"] = "n";
                            }                                                      
                    
                            var currentVariationId = $(\'input[name="variation_id"]\').val();                                              
                            if (!currentVariationId || currentVariationId == "0") {                               
                                window.dispatchEvent(new CustomEvent("customily-product-outofstock"));                               
                            }
                            
                            function handleVariantionChange() {
                                var oldVariationId = currentVariationId;
                                currentVariationId = $(\'input[name="variation_id"]\').val();
                                if (currentVariationId != oldVariationId) {                                    
                                    if (
                                        !currentVariationId ||
                                        (dataObj[currentVariationId] && dataObj[currentVariationId] === "n")
                                    ) {
                                        if (changeButtonText === "y") {
                                            $(button).html("' . esc_html($out_of_stock_text) . '");
                                        }
                                        window.dispatchEvent(new CustomEvent("customily-product-outofstock", {
                                            detail: {
                                                variationId: currentVariationId
                                            }
                                        }));
                                    } else {
                                        if (changeButtonText === "y") {
                                            $(button).html(buttonText);
                                        }
                                        window.dispatchEvent(new CustomEvent("customily-product-instock", {
                                            detail: {
                                                variationId: currentVariationId
                                            }
                                        }));
                                    }
                                }
                            }
                    
                            $(document).on("change", ".variations select", function (ev) {
                                const timeout = window.customilyWordpress?.variations?.handlerTimeout || 250;
                                setTimeout(function () {
                                    handleVariantionChange(ev);
                                }, timeout);                                
                            });
                        });
                    });
                </script>';
        }
    }

    /**
     * Add customily data to cart item
     * @param $cart_item_data
     * @param $product_id
     * @return mixed
     */
    public function wc_customily_add_custom_data_to_cart_item($cart_item_data, $product_id) {
        if (isset($_POST['properties'])) {
            $properties = wp_unslash($_POST['properties']);
            $session_id = sanitize_text_field($properties['_customily-session-id']);
            $missing_images = $_FILES;
            if (isset($session_id) && isset($missing_images)) {
                $results = Customily_Options_Utils::save_missing_images($missing_images, $session_id);
                foreach ($results as $result) {
                    if (isset($result['property']) && isset($result['url'])) {
                        $properties[$result['property']] = $result['url'];
                    }
                }
            }

            if (!$this->skip_add_properties_to_item($product_id, $properties, $cart_item_data)) {
                $this->app_settings = isset($this->app_settings['enabled'])
                    ? $this->app_settings
                    : $this->api_client->get_shop_settings();

                if (isset($this->app_settings) && $this->app_settings['enabled']) {
                    foreach ($properties as $label => $value) {
                        $key = 'property__' . str_replace(' ', CUSTOMILY_SPACE_REPLACEMENT, $label);

                        if (substr_count($value, 'cdn.customily.com') > 0 && substr_count($value, 'https://') == 0) {
                            $value = "https://{$value}";
                        }

                        $cart_item_data[$key] = array(
                            'label' => $label,
                            'value' => str_replace('\\', '', $value)
                        );
                    }
                }
            }
        }

        $cart_item_data['property___customily-added-date'] = [
            'label' => 'Added date',
            'value' => gmdate('Y-m-d H:i:s')
        ];

        return $cart_item_data;
    }

    /**
     * Display cart item price
     * @param $html
     * @param $cart_item
     * @param $cart_item_key
     */
    public function wc_customily_display_cart_item_price_html($html, $cart_item, $cart_item_key)
    {
        if (
            !class_exists('WC_Dynamic_Pricing_Compatibility') ||
            !isset($cart_item['discounts'])
        ) {
            return $html;
        }

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($cart_item['product_id']);
        if ($this->can_load_customily === 'n') {
            return $html;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $product = $cart_item['data'];
            $discounts = $cart_item['discounts'];

            if (function_exists('get_product')) {
                if (isset($cart_item['is_deposit']) && $cart_item['is_deposit']) {
                    $adjusted_price = $discounts['price_adjusted'];
                } else {
                    //Just use the price from the product, it has already been set during cart_loaded_from_session.
                    $adjusted_price = $product->get_price();
                }

                $adjusted_price = get_option('woocommerce_tax_display_cart') == 'excl'
                                ? wc_get_price_excluding_tax($product, array('price' => $adjusted_price, 'qty' => 1))
                                : wc_get_price_including_tax($product, array('price' => $adjusted_price, 'qty' => 1));
            } else {
                if (get_option('woocommerce_display_cart_prices_excluding_tax') == 'yes') {
                    $adjusted_price = wc_get_price_excluding_tax($product);
                } else {
                    $adjusted_price = $product->get_price();
                }
            }

            $base_price = floatval($discounts['display_price']);

            if (isset($cart_item['property___customily-personalization-price'])) {
                $base_price += $cart_item['property___customily-personalization-price']['value'];
            }

            if (!empty($adjusted_price) || $adjusted_price === 0) {
                if (apply_filters('wc_dynamic_pricing_use_discount_format', true)) {
                    $html = '<del style="display: block; margin: auto 15px">' .
                                WC_Dynamic_Pricing_Compatibility::wc_price($base_price) .
                            '</del>' .
                            '<ins style="margin: auto 15px">' .
                                WC_Dynamic_Pricing_Compatibility::wc_price($adjusted_price) .
                            '</ins>';
                } else {
                    $html = '<span class="amount">' .
                                WC_Dynamic_Pricing_Compatibility::wc_price($adjusted_price) .
                            '</span>';
                }
            }
        }

        return $html;
    }

    /**
     * Adjust cart item price
     *
     * @param WC_Cart $cart_obj
     */
    public function wc_customily_adjust_cart_items_price($cart_obj) {
        if (defined('CUSTOMILY_OPTIONS_CART_ITEMS_PRICE_ADJUSTED')) {
            return;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            foreach($cart_obj->get_cart() as $item) {
                if (isset($item['property___customily-personalization-price'])) {
                    $item['data']->set_price(
                        floatval($item['data']->get_price()) +
                        $item['property___customily-personalization-price']['value']
                    );
                }
            }
            define('CUSTOMILY_OPTIONS_CART_ITEMS_PRICE_ADJUSTED', true);
        }
    }

    /**
     * Show custom thumbnail on cart
     *
     * @param $product_image
     * @param $cart_item
     * @return string
     */
    public function wc_customily_cart_item_thumbnail($product_image, $cart_item) {
        if (
            !isset($cart_item['property___customily-thumb-base64']) &&
            !isset($cart_item['property[_customily-thumb]']) &&
            !isset($cart_item['property___customily-thumb'])
        ) {
            return $product_image;
        }

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($cart_item['product_id']);
        if ($this->can_load_customily === 'n') {
            return $product_image;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $show_thumbnail_link = $this->get_setting_value(
                'cl_show_thumbnail_link',
                $this->app_settings
            );

            $replace_thumbnail_on_cart = $this->get_setting_value(
                'replace_thumbnail_on_cart',
                $this->app_settings
            );

            if (!$show_thumbnail_link || !$replace_thumbnail_on_cart) {
                return $product_image;
            }

            $property_thumbnail_data_src = null;

            if (isset($cart_item['property[_customily-thumb]'])) {
                $property_thumbnail_data_src = $cart_item['property[_customily-thumb]']['value'];
            } else if (isset($cart_item['property___customily-thumb'])) {
                $property_thumbnail_data_src = $cart_item['property___customily-thumb']['value'];
            }

            $property_thumbnail_src = $cart_item['property___customily-thumb-base64'];
            $property_thumbnail_src = isset($property_thumbnail_src)
                                    ? "data:image/png;base64,{$property_thumbnail_src['value']}"
                                    : $property_thumbnail_data_src;

            if (isset($cart_item['property[_customily-thumb-id]'])) {
                $property_thumbnail_id = $cart_item['property[_customily-thumb-id]'];
            } else if (isset($cart_item['property___customily-thumb-id'])) {
                $property_thumbnail_id = $cart_item['property___customily-thumb-id'];
            }

            if (isset($cart_item['property[_customily-customization-url]'])) {
                $property_personalization_url = $cart_item['property[_customily-customization-url]'];
            } else if (isset($cart_item['property___customily-customization-url'])) {
                $property_personalization_url = $cart_item['property___customily-customization-url'];
            }

            if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                Customily_Options_Utils::add_log("Customily Wordpress: Showing custom thumbnail on cart: {$property_thumbnail_data_src}");
            }

            $data_thumb_id = !empty($property_thumbnail_id) ? 'data-thumb-id="' . $property_thumbnail_id['value'] . '"' : '';
            $product_image = "<img class=\"customily-thumb\" {$data_thumb_id} src=\"{$property_thumbnail_src}\""
                           . " data-thumb-src=\"{$property_thumbnail_data_src}\" style=\"object-fit: contain;\" />";

            if (isset($property_personalization_url['value']) && $this->app_settings['enableSavePersonalization']) {
                $product_image = '<a class="customily-customization-url" href="' . $property_personalization_url['value'] . '">'
                                    . $product_image .
                                '</a>';
            }
        }
        return $product_image;
    }

    /**
     * Show custom thumbnail on cart
     *
     * @param $product_images
     * @param $cart_item
     * @param $cart_item_key
     * @return mixed|string
     */
    public function wc_customily_store_api_cart_item_images($product_images, $cart_item, $cart_item_key) {
        if (
            !isset($cart_item['property___customily-thumb-base64']) &&
            !isset($cart_item['property[_customily-thumb]']) &&
            !isset($cart_item['property___customily-thumb'])
        ) {
            return $product_images;
        }

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($cart_item['product_id']);
        if ($this->can_load_customily === 'n') {
            return $product_images;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $show_thumbnail_link = $this->get_setting_value(
                'cl_show_thumbnail_link',
                $this->app_settings
            );

            $replace_thumbnail_on_cart = $this->get_setting_value(
                'replace_thumbnail_on_cart',
                $this->app_settings
            );

            if (!$show_thumbnail_link || !$replace_thumbnail_on_cart) {
                return $product_images;
            }

            $property_thumbnail_src = null;

            if (isset($cart_item['property[_customily-thumb]'])) {
                $property_thumbnail_src = $cart_item['property[_customily-thumb]']['value'];
            } else if (isset($cart_item['property___customily-thumb'])) {
                $property_thumbnail_src = $cart_item['property___customily-thumb']['value'];
            }

            if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                Customily_Options_Utils::add_log("Customily Wordpress: Showing custom thumbnail on cart: {$property_thumbnail_src}");
            }

            $product_images = [
                (object)[
                    'id'        => 0,
                    'src'       => $property_thumbnail_src,
                    'thumbnail' => $property_thumbnail_src,
                    'srcset'    => '',
                    'sizes'     => '',
                    'name'      => 'Personalized product image',
                    'alt'       => 'Personalized product image',
                ]
            ];
        }

        return $product_images;
    }

    /**
     * Display customily data in the cart.
     *
     * @param array $item_data
     * @param array $cart_item
     * @return array
     */
    public function wc_customily_show_custom_data_in_cart($item_data, $cart_item) {
        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($cart_item['product_id']);
        if ($this->can_load_customily === 'n') {
            return $item_data;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            /*
                file-numbers property contains the file numbers separated by commas.
                When a product has multiple files, the file numbers are used as a suffix to property:
                personalization-id, production-file-url, eps-name and eps-placeholder.
                For file number 1 no suffix is used, this is to maintain backward compatibility.
                Example: If the product has 2 files, we will have these properties:
                - _customily-personalization-id
                - _customily-personalization-id-2
                - _customily-production-url
                - _customily-production-url-2
                - _customily-eps-name
                - _customily-eps-name-2
                - _customily-eps-placeholder
                - _customily-eps-placeholder-2
            */

            if (isset($cart_item['property___customily-file-numbers'])) {
                $file_numbers = $cart_item['property___customily-file-numbers'];
            }

            $multi_file_properties = !empty($file_numbers) ? $this->get_multi_files_properties($file_numbers['value']) : [];

            $show_personalization = $this->get_setting_value(
                'cl_show_personalization',
                $this->app_settings
            );
            $show_preview_link = $this->get_setting_value(
                'cl_show_preview_link',
                $this->app_settings
            );
            $show_personalization_price = get_option('cl_show_personalization_price');

            foreach ($cart_item as $key => $item) {
                if (
                    (
                        substr_count($key, 'property__') > 0 ||
                        substr_count($key, 'cl_input') > 0
                    ) &&
                    substr_count($key, '_customily-id-') == 0 &&
                    !in_array($key, PROPERTIES_SEARCH_ARRAY) &&
                    // Also take into account properties used in multi files to which the file number is added as a suffix.
                    !in_array($key, $multi_file_properties)
                ) {
                    if (!empty($item['label']) && !empty($item['value'])) {
                        if (
                            !$show_preview_link &&
                            substr_count($key, '_customily-preview') > 0
                        ) {
                            continue;
                        }

                        if (
                            !$show_personalization &&
                            substr_count($key, '_customily-preview') == 0
                        ) {
                            continue;
                        }

                        if (
                            $show_personalization_price == 'n' &&
                            substr_count($key, '_customily-personalization-price') > 0
                        ) {
                            continue;
                        }

                        $prop = $this->get_property($key, $item, false, true);
                        if (isset($prop)) {
                            $item_data[] = array(
                                'key' => $prop['label'],
                                'value' => $prop['value']
                            );
                        }
                    }
                }
            }
        }

        return $item_data;
    }

    /**
     * Load cart script
     *
     * @return void
     */
    public function wc_customily_load_cart_script() {
        wp_enqueue_script(
            esc_html(CUSTOMILY_OPTIONS_PLUGIN_NAME),
            plugin_dir_url(__FILE__) . 'scripts/customily-options-public-cart.js',
            array(),
            esc_html(CUSTOMILY_OPTIONS_PLUGIN_VERSION),
            true
        );
    }

    /**
     * Show custom thumbnail on checkout
     *
     * @param $name
     * @param $cart_item
     * @param $cart_item_key
     * @return mixed|string
     */
    public function wc_customily_show_custom_thumbnail_on_checkout($name, $cart_item, $cart_item_key) {
        if (
            !is_checkout() ||
            (
                !isset($cart_item['property[_customily-thumb]']) &&
                !isset($cart_item['property___customily-thumb'])
            )
        ) {
            return $name;
        }

        $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($cart_item['product_id']);
        if ($this->can_load_customily === 'n') {
            return $name;
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $show_thumbnail_link = $this->get_setting_value(
                'cl_show_thumbnail_link',
                $this->app_settings
            );
            $replace_thumbnail_on_checkout = $this->get_setting_value(
                'cl_thumbnail_on_checkout',
                $this->app_settings
            );

            if (!$show_thumbnail_link || !$replace_thumbnail_on_checkout) {
                return $name;
            }

            // if $replace_thumbnail_on_checkout == false add thumbnail on item name
            $product = apply_filters(
                'woocommerce_cart_item_product',
                $cart_item['data'],
                $cart_item,
                $cart_item_key
            );

            $property = $cart_item['property[_customily-thumb]'] ?? $cart_item['property___customily-thumb'];
            $property = $property['value'];

            if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                Customily_Options_Utils::add_log("Customily Wordpress: Showing custom thumbnail on checkout: {$property}");
            }

            $thumbnail = "<img src=\"{$property}\" style=\"object-fit: contain;\"/>";
            $product_name = $product->get_name();

            if (function_exists('t4t_product_image_on_checkout')) {
                $image = '<div class="t4t-product-checkout-item"><div class="t4t-product-image">' . $thumbnail . '</div>';
                $product_name = '<span class="t4t-product-name">' . $product_name . '</span></div>';
            } else {
                echo '<style>
                    .cl-custom-thumbnail {
                        width: 52px;
                        height: 45px;
                        display: inline-block !important;
                        padding-right: 7px;
                        vertical-align: middle;
                    }
                    .ts-product-image { display: none !important; }
                </style>';

                $image = '<div class="cl-custom-thumbnail">' . $thumbnail . '</div>';
            }

            $name = "{$image}{$product_name}";
        }
        return $name;
    }

    /**
     * Add customily data to order item details.
     *
     * @param WC_Order_Item_Product $order_item
     * @param string                $item_key
     * @param array                 $values
     * @param WC_Order              $order
     */
    public function wc_customily_add_custom_data_to_order_item($order_item, $item_key, $values, $order) {
        $order_id = $order->save();

        require_once plugin_dir_path(dirname(__FILE__)) . 'includes/customily-options-device-info.php';

        if (!empty($_SERVER['HTTP_USER_AGENT'])) {
            $device_info = new Customily_Options_Device_Info();
            $device_all = $device_info->getAll(sanitize_text_field(wp_unslash($_SERVER['HTTP_USER_AGENT'])));

            if (!$order->meta_exists('Mobile') && !empty($device_all['device_type'])) {
                $mobile = sanitize_text_field(wp_unslash($device_all['device_type']));
                $order->add_meta_data(__('Mobile', 'customily-v2'), $mobile);
                if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                    Customily_Options_Utils::add_log("Customily Wordpress: order {$order_id}, device type {$mobile}");
                }
            }

            if (!$order->meta_exists('OS') && !empty($device_all['os_title'])) {
                $os = sanitize_text_field(wp_unslash($device_all['os_title']));
                $order->add_meta_data(__('OS', 'customily-v2'), $os);
                if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                    Customily_Options_Utils::add_log("Customily Wordpress: order {$order_id}, OS {$os}");
                }
            }

            if (!$order->meta_exists('Browser') && !empty($device_all['browser_title'])) {
                $browser = sanitize_text_field(wp_unslash($device_all['browser_title']));
                $order->add_meta_data(__('Browser', 'customily-v2'), $browser);
                if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                    Customily_Options_Utils::add_log("Customily Wordpress: order {$order_id}, browser {$browser}");
                }
            }
        }

        if (isset($values['property___customily-template-id'])) {
            $template_id = $values['property___customily-template-id'];
        } else if (isset($values['cl_template_id'])) {
            $template_id = [
                'value' => $values['cl_template_id']
            ];
        }

        if (!empty($template_id)) {
            $order->add_meta_data("template_id_{$item_key}", $template_id['value']);
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {          
            /*
                file-numbers property contains the file numbers separated by commas.
                When a product has multiple files, the file numbers are used as a suffix to property:
                personalization-id, production-file-url, eps-name and eps-placeholder.
                For file number 1 no suffix is used, this is to maintain backward compatibility.
                Example: If the product has 2 files, we will have these properties:
                - _customily-personalization-id
                - _customily-personalization-id-2
                - _customily-production-url
                - _customily-production-url-2
                - _customily-eps-name
                - _customily-eps-name-2
                - _customily-eps-placeholder
                - _customily-eps-placeholder-2
            */

            if (isset($values['property___customily-file-numbers'])) {
                $file_numbers = $values['property___customily-file-numbers'];
            }

            // By default, if the file-number property doesn't exist, initialize the array with file number '1',
            // this way no suffix will be added.
            $file_numbers_array = ['1'];
            if (!empty($file_numbers)) {
                $order->add_meta_data("file_numbers_{$item_key}", $file_numbers['value']);
                $file_numbers_array = explode(',', $file_numbers['value']);
            }

            $multi_file_properties = [];

            foreach ($file_numbers_array as $file_number) {
                $file_number = (int) $file_number;
                $customily_suffix = '';
                $order_meta_suffix = '';

                //If the file number is 1, no suffix is used, this was done to maintain backward compatibility.
                if ($file_number > 1) {
                    $customily_suffix = "-{$file_number}";
                    $order_meta_suffix = "{$file_number}_";
                }

                if (isset($values["property[_customily-personalization-id{$customily_suffix}]"])) {
                    $personalization_id = $values["property[_customily-personalization-id{$customily_suffix}]"];
                } else if (isset($values["property___customily-personalization-id{$customily_suffix}"])) {
                    $personalization_id = $values["property___customily-personalization-id{$customily_suffix}"];
                } else if (isset($values['cl_personalization_id'])) {
                    $personalization_id = [
                        'value' => $values['cl_personalization_id']
                    ];
                }
    
                if (isset($values["property[_customily-production-url{$customily_suffix}]"])) {
                    $production_url = $values["property[_customily-production-url{$customily_suffix}]"];
                } else if (isset($values["property___customily-production-url{$customily_suffix}"])) {
                    $production_url = $values["property___customily-production-url{$customily_suffix}"];
                }

                if (isset($values["property___customily-eps-name{$customily_suffix}"])) {
                    $eps_name = $values["property___customily-eps-name{$customily_suffix}"]['value'];
                    $order->add_meta_data(
                        "eps_name_{$order_meta_suffix}{$item_key}",
                        $eps_name
                    );
                }

                if (isset($values["property___customily-eps-placeholder{$customily_suffix}"])) {
                    $order->add_meta_data(
                        "eps_placeholder_{$order_meta_suffix}{$item_key}",
                        $values["property___customily-eps-placeholder{$customily_suffix}"]['value']
                    );
                }

                if (
                    !empty($personalization_id) &&
                    !empty($production_url)
                ) {
                    $personalization_id = str_replace('_customily-id-', '', $personalization_id['value']);
                    $order->add_meta_data("personalization_id_{$order_meta_suffix}{$item_key}", $personalization_id);
    
                    if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                        Customily_Options_Utils::add_log("Customily Wordpress: Personalization id: {$personalization_id}");
                    }
    
                    $file_url = $production_url['value'];
                    if (isset($this->app_settings['renameProdFile']) && $this->app_settings['renameProdFile']) {
                        $old_file_name = pathinfo($file_url, PATHINFO_FILENAME);
                        $is_multi_file = count($file_numbers_array) > 1;
                        $new_file_name = $this->get_production_file_name(
                            $order,
                            $order_item,
                            $item_key,
                            $is_multi_file,
                            $file_number,
                            $eps_name ?? null
                        );
                        $file_url = str_replace($old_file_name, $new_file_name, $file_url);
    
                        if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                            Customily_Options_Utils::add_log(
                                "Customily Wordpress:" .
                                " Renaming production file => product name: {$order_item->get_name()}," .
                                " personalization id: {$personalization_id}, order number: {$order->get_order_number()}," .
                                " oldName: {$old_file_name}, newName: {$new_file_name}"
                            );
                        }
    
                        $order->add_meta_data("production_file_renamed_{$order_meta_suffix}{$item_key}", true);
                    }
    
                    $order->add_meta_data("production_file_{$order_meta_suffix}{$item_key}", $file_url);
    
                    if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                        Customily_Options_Utils::add_log(
                            "Customily Wordpress: Production file => product name: {$order_item->get_name()}," .
                            " personalization id: {$personalization_id}, order number: {$order->get_order_number()}," .
                            " url: {$file_url}"
                        );
                    }
                }

                if ($file_number > 1) {
                    $file_properties = $this->get_file_properties($file_number);
                    $multi_file_properties = array_merge($multi_file_properties, $file_properties);
                }
            }

            $order_item->add_meta_data('_cart_item_key', $item_key);

            foreach ($values as $key => $item) {
                if (
                    (
                        substr_count($key, 'property__') > 0 ||
                        substr_count($key, 'cl_input') > 0
                    ) &&
                    substr_count($key, '_customily-id-') == 0 &&
                    !in_array($key, PROPERTIES_SEARCH_ARRAY) &&
                    // Also take into account properties used in multi files to which the file number is added as a suffix.
                    !in_array($key, $multi_file_properties)
                ) {
                    if (isset($item['label']) && !empty($item['value'])) {
                        $prop = $this->get_property($key, $item);
                        if (isset($prop)) {
                            $order_item->add_meta_data(esc_html($prop['label']), $prop['value']);
                            if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                                Customily_Options_Utils::add_log(
                                    "Customily Wordpress: Option label: {$prop['label']} => value: {$prop['value']}"
                                );
                            }
                        }
                    }
                }
            }

            if (isset($values['property[_customily-thumb]'])) {
                $thumbnail = $values['property[_customily-thumb]'];
            } else if (isset($values['property___customily-thumb'])) {
                $thumbnail = $values['property___customily-thumb'];
            }

            if (!empty($thumbnail)) {
                $order_item->add_meta_data(
                    __('_customily-thumb', 'customily-v2'),
                    $thumbnail['value']
                );
            }

            $customization_url = $this->get_customization_url_property($values);

            if (!empty($customization_url)) {
                $order_item->add_meta_data(
                    __('_customily-customization-url', 'customily-v2'),
                    $customization_url
                );
            }

            $order->update_meta_data(CUSTOMILY_OPTIONS_ORDER_CREATION_NOTIFIED, 'n');
        }

        if (isset($values['property___customily-woocommerce-product-id'])) {
            $product_id = $values['property___customily-woocommerce-product-id'];
        }

        if (!empty($product_id)) {
            $order_item->add_meta_data(
                __('Woocommerce Product Id', 'customily-v2'),
                $product_id['value']
            );
        }

        if (isset($values['property[_customily-added-date]'])) {
            $added_date = $values['property[_customily-added-date]'];
        } else if (isset($values['property___customily-added-date'])) {
            $added_date = $values['property___customily-added-date'];
        } else if (isset($values['cl_input-added-date'])) {
            $added_date = [
                'value' => $values['cl_input-added-date']
            ];
        }

        if (!empty($added_date)) {
            $order_item->add_meta_data(
                __('Added date', 'customily-v2'),
                $added_date['value']
            );
        }
    }

    /**
     * Hook order payment completed to notified order creation to shopify app.
     *
     * @param $order_id
     */
    public function wc_customily_woocommerce_payment_complete_hook($order_id) {
        if (!$order_id) {
            return;
        }

        Customily_Options_Utils::add_log("Customily Wordpress: order payment completed for order '{$order_id}'");
        $this->wc_customily_notify_order_creation($order_id);
    }

    /**
     * Notify order creation and handle printful/printify.
     * @param $order_id
     * @param bool $from_api
     * @return false|string
     */
    public function wc_customily_notify_order_creation($order_id, $from_api = false) {
        if (!$order_id) {
            return false;
        }

        $wc_order = wc_get_order($order_id);
        if (empty($wc_order)) {
            return 'not_found';
        }

        $already_notified = 'n';
        if ($wc_order->meta_exists(CUSTOMILY_OPTIONS_ORDER_CREATION_NOTIFIED)) {
            $already_notified = $wc_order->get_meta(CUSTOMILY_OPTIONS_ORDER_CREATION_NOTIFIED);
        }

        if ($already_notified == 'y' && !$from_api) {
            return 'already_notified';
        }

        $notified = 'y';
        $orders_db = new Customily_Options_Orders_DB();
        $order = $orders_db->get_order_by_id($order_id, true, true);

        if ($order['personalized_count'] > 0) {
            $this->app_settings = isset($this->app_settings['enabled'])
                ? $this->app_settings
                : $this->api_client->get_shop_settings();

            if (isset($this->app_settings) && $this->app_settings['enabled']) {
                Customily_Options_Utils::add_log(
                    "Customily Wordpress: Notifying order creation for order '{$order_id}'," .
                    " personalized items: {$order['personalized_count']}."
                );

                $notified = $this->api_client->order_created($order);
                if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                    $notified == 'y' ?
                        Customily_Options_Utils::add_log(
                            "Customily Wordpress: Order '{$order_id}' creation notified," .
                            " personalized items: {$order['personalized_count']}."
                        ) : Customily_Options_Utils::add_log(
                            "Customily Wordpress: Error when trying to notify order '{$order_id}' creation."
                    );
                }
            } else {
                $notified = 'customily_disabled';
            }
            if ($notified != 'y') {
                require_once plugin_dir_path(dirname(__FILE__)) . 'admin/db/customily-options-orders-notifications-db.php';
                $db = new Customily_Options_Orders_Notifications_DB();
                if ($db->create_or_update_order_notification($order)) {
                    Customily_Options_Utils::add_log("Customily Wordpress: Order notification created/updated for order {$order_id}");
                }
            }
        }

        // Flag the action as done (to avoid repetitions on reload for example)
        $wc_order->update_meta_data(CUSTOMILY_OPTIONS_ORDER_CREATION_NOTIFIED, $notified);
        $wc_order->save();
        return $notified;
    }

    /**
     * Show custom thumbnail on order pay
     *
     * @param $name
     * @param WC_Order_Item $item
     * @return mixed|string
     * @throws Exception
     */
    public function wc_customily_woocommerce_order_item_name($name, $item) {
        $thumbnail_label = '_customily-thumb';

        if (!is_checkout() || !$item->meta_exists($thumbnail_label)) {
            return $name;
        }

        $product = $this->get_product_by_order_item($item);
        if ($product) {
            $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($product->get_id());
            if ($this->can_load_customily === 'n') {
                return $name;
            }
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $show_thumbnail_link = $this->get_setting_value(
                'cl_show_thumbnail_link',
                $this->app_settings
            );
            $replace_thumbnail_on_order_pay = $this->get_setting_value(
                'cl_thumbnail_on_order_pay',
                $this->app_settings
            );

            if (!$show_thumbnail_link || $replace_thumbnail_on_order_pay) {
                return $name;
            }

            // if $replace_thumbnail_on_order_pay == false add thumbnail on item name
            $item_name = $item->get_name();
            $link = Customily_Options_Utils::extract_url_from_link($item->get_meta($thumbnail_label));

            if (isset($this->app_settings['debugMode']) && $this->app_settings['debugMode']) {
                Customily_Options_Utils::add_log("Customily Wordpress: Showing custom thumbnail on order pay: {$link}");
            }

            $thumbnail = "<img src=\"{$link}\" style=\"object-fit: contain;\"/>";

            echo '<style>
                    .cl-custom-thumbnail {
                        width: 52px;
                        height: 45px;
                        display: inline-block !important;
                        padding-right: 7px;
                        vertical-align: middle;
                    }
                    .ts-product-image { display: none !important; }
                  </style>';

            $image = '<div class="cl-custom-thumbnail">'
                        . $thumbnail .
                    '</div>';

            /* Prepend image to name and return it */
            $name = "{$image}{$item_name}";
        }
        return $name;
    }

    /**
     * Get order item formatted meta_data
     * @param array $formatted_meta
     * @param WC_Order_Item $item
     * @return array
     */
    public function wc_customily_order_item_get_formatted_meta_data($formatted_meta, $item)
    {
        $product = $this->get_product_by_order_item($item);
        $is_admin_view = is_admin();
        if ($product) {
            $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($product->get_id());
            if ($this->can_load_customily === 'n') {
                return $formatted_meta;
            }
        }

        $this->app_settings = isset($this->app_settings['enabled'])
            ? $this->app_settings
            : $this->api_client->get_shop_settings();

        if (isset($this->app_settings) && $this->app_settings['enabled']) {
            $added_date_label = 'Added date';
            $preview_label = $this->get_setting_value('preview_link_label', $this->app_settings);
            $thumbnail_label = '_customily-thumb';
            $customization_url_label = '_customily-customization-url';
            $woocommerce_product_id_label = 'Woocommerce Product Id';

            $show_personalization = $this->get_setting_value(
                'cl_show_personalization',
                $this->app_settings
            );
            $show_preview_link = $this->get_setting_value(
                'cl_show_preview_link',
                $this->app_settings
            );

            // hide preview link
            $this->unset_formatted_meta($formatted_meta, '_customily-preview');
            $preview_url = $item->get_meta('_customily-preview');

            if ($show_preview_link || $is_admin_view) {
                if (!empty($preview_url)) {
                    $formatted_meta[99999997] = (object)array(
                        'key' => $preview_label,
                        'value' => $preview_url,
                        'display_key' => $preview_label,
                        'display_value' => $preview_url
                    );
                }
            }

            if ($show_personalization || $is_admin_view) {
                // hide cart item key
                $this->unset_formatted_meta($formatted_meta, '_cart_item_key');
                // hide added to cart date
                $this->unset_formatted_meta($formatted_meta, $added_date_label);
                // hide thumbnail link
                $this->unset_formatted_meta($formatted_meta, $thumbnail_label);
                // hide woocommerce product id
                $this->unset_formatted_meta($formatted_meta, $woocommerce_product_id_label);
                // hide thumbnail base64
                $this->unset_formatted_meta($formatted_meta, '_customily-thumb-base64');
            } else {
                $formatted_meta = array();
                $meta_data = $item->get_meta_data();
                $item_name = $item->get_name();

                foreach ($meta_data as $meta) {
                    if (
                        empty($meta->id) ||
                        $meta->value == '' ||
                        !is_scalar($meta->value) ||
                        in_array($meta->key, array(
                            '_cart_item_key',
                            '_reduced_stock',
                            $added_date_label,
                            $thumbnail_label
                        ))
                    ) {
                        continue;
                    }

                    if (!$show_preview_link && $meta->key == $preview_label) {
                        continue;
                    }

                    $meta->key = rawurldecode((string)$meta->key);
                    $meta->value = rawurldecode((string)$meta->value);
                    $attribute_key = str_replace('attribute_', '', $meta->key);
                    $display_key = wc_attribute_label($attribute_key, $product);
                    $display_value = wp_kses_post($meta->value);
                    $taxonomy_exists = taxonomy_exists($attribute_key);

                    if ($taxonomy_exists) {
                        $term = get_term_by('slug', $meta->value, $attribute_key);
                        if (!is_wp_error($term) && is_object($term) && $term->name) {
                            $display_value = $term->name;
                        }
                    }

                    // Skip items with values already in the product details area of the product name.
                    if (
                        $product &&
                        (
                            (
                                $product->is_type('variation') &&
                                wc_is_attribute_in_product_name($display_value, $item_name)
                            ) ||
                            (
                                (!$taxonomy_exists || !$product->is_type('variation')) &&
                                $meta->key != $preview_label
                            )
                        )
                    ) {
                        continue;
                    }

                    $formatted_meta[$meta->id] = (object)array(
                        'key' => $meta->key,
                        'value' => $meta->value,
                        'display_key' => apply_filters(
                            'woocommerce_order_item_display_meta_key',
                            $display_key,
                            $meta,
                            $item
                        ),
                        'display_value' => wpautop(
                            make_clickable(
                                apply_filters(
                                    'woocommerce_order_item_display_meta_value',
                                    $display_value,
                                    $meta,
                                    $item
                                )
                            )
                        ),
                    );
                }
            }

            if ($is_admin_view) {
                $order = $item->get_order();
                $customization_url = $item->get_meta($customization_url_label);

                if (!empty($customization_url)) {
                    $this->unset_formatted_meta($formatted_meta, $customization_url_label);

                    $show_customization_url_as_text = get_option('cl_show_customization_url_as_text');
                    if ($show_customization_url_as_text != 'y' && substr_count($customization_url, 'href') == 0) {
                        $customization_url = '<a href="' . $customization_url . '" target="_blank">Link</a>';
                    }

                    $formatted_meta[99999990] = (object)array(
                        'key' => 'Customization Url',
                        'value' => $customization_url,
                        'display_key' => 'Customization Url',
                        'display_value' => $customization_url
                    );
                }

                $item_key = $item->get_meta('_cart_item_key');

                $file_numbers_array = ['1'];
                if ($order->meta_exists("file_numbers_{$item_key}")) {
                    $file_numbers = $order->get_meta("file_numbers_{$item_key}");
                    $file_numbers_array = explode(",", $file_numbers);
                }

                foreach ($file_numbers_array as $file_number) {
                    $file_number = (int) $file_number;
                    $order_meta_suffix = '';

                    // If the file number is 1, no suffix is used, this was done to maintain backward compatibility.
                    if ($file_number > 1) {
                        $order_meta_suffix = "{$file_number}_";
                    }

                    $file_url = $order->get_meta("production_file_{$order_meta_suffix}{$item_key}");
                    if (empty($file_url)) {
                        $file_url = $order->get_meta("Production File #{$item_key}");
                    }

                    $predicate = $file_number > 1 ? " {$file_number}" : '';

                    if (!empty($file_url)) {
                        $meta_key = "Production File{$predicate}";
                        $this->unset_formatted_meta($formatted_meta, $meta_key);
                        $meta_value = '<a href="' . $file_url . '" target="_blank">Link</a>';
                        $formatted_meta[99999990 + $file_number] = (object)array(
                            'key' => $meta_key,
                            'value' => $meta_value,
                            'display_key' => $meta_key,
                            'display_value' => $meta_value
                        );
                    }
                }
            }
        }
        return $formatted_meta;
    }

    /**
     * Get personalization property item
     *
     * @param $key
     * @param $item
     * @param $return_default_label
     * @param $skip_hidden
     * @return array
     */
    private function get_property($key, $item, $return_default_label = false, $skip_hidden = false) {
        $result = [
            'label' => $item['label'],
            'value' => $item['value']
        ];

        if (
            substr_count($key, '_customily-preview') > 0 ||
            substr_count($key, 'cl_custom_thumb') > 0
        ) {
            $show_preview_link = $this->get_setting_value(
                'cl_show_preview_link',
                $this->app_settings
            );
            if ($show_preview_link || !$skip_hidden) {
                $result = $this->get_preview_link_property($result, $return_default_label);
            } else {
                $result = null;
            }
        }

        if (
            substr_count($result['label'], '__uploaded__') > 0 ||
            substr_count($result['label'], 'Uploaded') > 0
        ) {
            $show_uploaded_link = $this->get_setting_value(
                'cl_show_uploaded_link',
                $this->app_settings
            );
            if ($show_uploaded_link == 'y' || !$skip_hidden) {
                $result['label'] = str_replace('__uploaded__', '', $result['label']);
                $result['label'] = str_replace('Uploaded', '', $result['label']);
                $value = $this->get_setting_value('uploaded_link_text', $this->app_settings);
                $result['value'] = '<a href="' . $result['value'] . '" target="_blank">' . $value . '</a>';
            } else {
                $result = null;
            }
        }

        if (substr_count($key, '_customily-personalization-price') > 0) {
            $result['label'] = $this->get_setting_value(
                'personalization_price_text',
                $this->app_settings
            );
            $result['value'] = get_woocommerce_currency_symbol() . number_format(
                (float)$result['value'],
                2,
                '.',
                ''
            );
        }

        return $result;
    }

    /**
     * Init custom strings (get custom strings from customJs setting)
     */
    private function init_custom_strings() {
        if (!$this->custom_strings) {
            $settings = $this->app_settings['customJs'] ?? $this->app_settings['settings']['customJs'];
            $start_position = strpos($settings, "#begin_json#");
            $end_position = strpos($settings, "#end_json#");

            if (!empty($start_position) && !empty($end_position)) {
                $start_position += 12;
                $str = substr($settings, $start_position, $end_position - $start_position);
                $this->custom_strings = json_decode($str, true);
            }
        }
    }

    /**
     * Get production file rename pattern
     *
     * @param WC_Order $order
     * @param WC_Order_Item_Product $item
     * @param string $cart_item_key
     * @param $is_multi_file
     * @param $file_number
     * @param $file_name
     * @return string
     */
    private function get_production_file_name($order, $item, $cart_item_key, $is_multi_file, $file_number, $file_name) {
        $result = '';
        $this->init_custom_strings();
        $pattern = $this->custom_strings['rename_pattern'] ?? '';

        if (empty($pattern) && isset($this->app_settings['renameProdFilePattern'])) {
            $pattern = $this->app_settings['renameProdFilePattern'];
        }

        if (strlen($pattern) > 0) {
            $pattern_str_array = explode('_', $pattern);
            $length = count($pattern_str_array);
            for ($i = 0; $i < $length - 1; $i++) {
                $result .= "{$this->get_pattern_item_value($pattern_str_array[$i], $order, $item, $cart_item_key, $file_number, $file_name)}_";
            }
            $result .= $this->get_pattern_item_value($pattern_str_array[$length - 1], $order, $item, $cart_item_key, $file_number, $file_name);
            
            if ($is_multi_file && !in_array('filenumber', $pattern_str_array)) {
                //If the product is multi file, the file number is added to avoid having repeated names.
                $result .= isset($file_number) ? "_number-{$file_number}" : '';
            }
        } else {
            $result = "{$order->get_order_number()}_{$cart_item_key}";
            if ($is_multi_file) {
                //If the product is multi file, the file number is added to avoid having repeated names.
                $result .= isset($file_number) ? "_number-{$file_number}" : '';
            }
        }

        return $result;
    }

    /**
     * Get production file pattern item
     *
     * @param $str string
     * @param WC_Order $order
     * @param WC_Order_Item_Product $item
     * @param string $cart_item_key
     * @param $file_number
     * @param $file_name
     * @return string
     */
    private function get_pattern_item_value($str, $order, $item, $cart_item_key, $file_number, $file_name) {
        $result = '';
        switch ($str) {
            case 'itemid':
                $result = $item->get_id();
                break;
            case 'productid':
                $result = $item->get_product_id();
                break;
            case 'sku':
                $result = Customily_Options_Utils::slugify(sanitize_title($item->get_product()->get_sku()));
                break;
            case 'productname':
                $result = Customily_Options_Utils::slugify(sanitize_title($item->get_name()));
                break;
            case 'ordernumber':
                $result = $order->get_order_number();
                break;
            case 'itemindex':
                $result = $order->get_meta(CUSTOMILY_OPTIONS_ORDERITEM_LAST_INDEX);
                $result = empty($result) ? 1 : (int)$result + 1;
                $order->update_meta_data(CUSTOMILY_OPTIONS_ORDERITEM_LAST_INDEX, $result);
                break;
            case 'itemqty':
                $result = $item->get_quantity();
                break;
            case 'itemkey':
                $result = $cart_item_key;
                break;
            case 'firstname':
                $first_name = $order->get_billing_first_name();
                $result = isset($first_name) ? Customily_Options_Utils::slugify(sanitize_title($first_name)) : '';
                break;
            case 'lastname':
                $last_name = $order->get_billing_last_name();
                $result = isset($last_name) ? Customily_Options_Utils::slugify(sanitize_title($last_name)) : '';
                break;
            case 'customer':
                $first_name = $order->get_billing_first_name();
                $last_name = $order->get_billing_last_name();
                if (isset($first_name) && isset($last_name)) {
                    $result = Customily_Options_Utils::slugify(sanitize_title("{$first_name} {$last_name}"));
                }
                break;
            case 'random':
                $result = wp_rand(10000, 99999);
                break;                
            case 'filenumber':
                $result = isset($file_number) ? "number-{$file_number}" : '';
                break;
            case 'filename':
                $result = isset($file_name) ? Customily_Options_Utils::slugify(sanitize_title($file_name)) : '';
                break;
            default:
                $result = Customily_Options_Utils::slugify($str);
                break;
        }
        return $result;
    }

    /**
     * Insert app initialize script
     */
    private function app_initialize_script() {
        require_once plugin_dir_path(__FILE__) . 'scripts/customily-options-public-app-init.php';
    }

    /**
     * Unset formatted meta element
     * @param $formatted_meta
     * @param $meta_key
     */
    private function unset_formatted_meta(&$formatted_meta, $meta_key) {
        $filter = function ($meta) use ($meta_key) {
            return $meta->key == $meta_key;
        };
        $data = array_filter($formatted_meta, $filter);
        if (!empty($data)) {
            $key = key($data);
            unset($formatted_meta[$key]);
        }
    }

    /**
     * Get setting value for backward compatibility
     * @param $setting_label
     * @param $settings
     * @return bool|mixed|string
     */
    private function get_setting_value($setting_label, $settings) {
        $result = '';
        if ($settings['enabled'] && isset($settings['woocommerce']['enabled'])) {
            switch ($setting_label) {
                case 'cl_show_personalization':
                    $result = $settings['woocommerce']['showPersonalization'] ?? true;
                    break;
                case 'cl_show_preview_link':
                    $result = $settings['woocommerce']['preview']['showLink'] ?? true;
                    break;
                case 'cl_show_thumbnail_link':
                    $result = $settings['woocommerce']['thumbnail']['showThumbnail'] ?? true;
                    break;
                case 'replace_thumbnail_on_cart':
                    $result = $settings['woocommerce']['thumbnail']['replaceCartThumbnail'] ?? true;
                    break;
                case 'cl_thumbnail_on_checkout':
                    $result = $settings['woocommerce']['thumbnail']['replaceCheckoutThumbnail'] ?? true;
                    break;
                case 'cl_thumbnail_on_order_pay':
                    $result = $settings['woocommerce']['thumbnail']['replaceOrderPayThumbnail'] ?? true;
                    break;
                case 'cl_show_uploaded_link':
                    $result = $settings['woocommerce']['uploaded']['showLink'] ?? true;
                    break;
                case 'cl_handle_product_stock_status':
                    $result = $settings['woocommerce']['productStockStatus']['handle'] ?? false;
                    break;
                case 'cl_change_button_text_by_product_stock_status':
                    $result = $settings['woocommerce']['productStockStatus']['changeButtonText'] ?? false;
                    break;
                case 'cl_prevent_duplicate_products':
                    // TODO use this setting to prevent add item to cart on page refresh
                    $result = $settings['woocommerce']['preventDuplicateProducts'] ?? true;
                    break;
                case 'preview_link_label':
                    $result = $settings['woocommerce']['preview']['linkLabel'] ?? 'Preview';
                    break;
                case 'preview_link_text':
                    $result = $settings['woocommerce']['preview']['linkText'] ?? 'Link';
                    break;
                case 'uploaded_link_text':
                    $result = $settings['woocommerce']['uploaded']['linkText'] ?? 'Link';
                    break;
                case 'add_to_cart_text':
                    $result = $settings['woocommerce']['productStockStatus']['buttonTextValue'] ?? 'Out of Stock';
                    break;
                case 'personalization_price_enabled':
                    $result = $settings['woocommerce']['optionsPrices']['enabled'] ?? false;
                    break;
                case 'personalization_price_text':
                    $result = $settings['woocommerce']['optionsPrices']['personalizationPriceText'] ?? 'Personalization Price';
                    break;
            }
        } else {
            $this->init_custom_strings();

            switch ($setting_label) {
                case 'preview_link_label':
                    $result = $this->custom_strings['preview'] ?? 'Preview';
                    break;
                case 'preview_link_text':
                case 'uploaded_link_text':
                    $result = $this->custom_strings['link'] ?? 'Link';
                    break;
                case 'replace_thumbnail_on_cart':
                    $result = true;
                    break;
                case 'add_to_cart_text':
                    $result = $this->custom_strings['out_of_stock'] ?? 'Out of Stock';
                    break;
                case 'personalization_price_enabled':
                    $result = false;
                    break;
                case 'personalization_price_text':
                    $result = $this->custom_strings['personalization_price_text'] ?? 'Personalization Price';
                    break;
                default:
                    $result = get_option($setting_label) == 'y';
                    break;
            }
        }

        return $result;
    }

    /**
     * Get product categories
     * @param $product_id
     * @return array|false|string[]
     */
    private function get_product_categories($product_id) {
        $terms = get_the_terms($product_id, 'product_cat');
        if (!$terms) {
            return false;
        }
        return array_map(function ($el) {
            return $el->name;
        }, $terms);
    }

    /**
     * Check if product has a category
     * @param $product_id
     * @param $category_name
     * @return bool
     */
    private function product_has_category($product_id, $category_name) {
        $categories = $this->get_product_categories($product_id);
        if (!$categories) {
            return false;
        }
        foreach ($categories as $category) {
            if (strtolower($category) === strtolower($category_name)) {
                return true;
            }
        }
        return false;
    }

    /**
     * Can load customily
     * @param $product_id
     * @return string
     */
    private function can_load_customily($product_id) {
        if (get_option('cl_check_products_connected') === 'y') {
            $category = get_option('cl_products_category');
            if (empty($category)) {
                $category = 'Customily';
            }
            if (!$this->product_has_category($product_id, $category)) {
                return 'n';
            }
        }
        return 'y';
    }

    /**
     * Get product by order item
     * @param WC_Order_Item $item
     * @return false|mixed
     */
    private function get_product_by_order_item($item) {
        return is_callable(array($item, 'get_product')) ? $item->get_product() : false;
    }

    /**
     * Find property value by label
     * @param $label
     * @param $properties
     * @return false|mixed
     */
    private function find_property($label, $properties) {
        foreach ($properties as $key => $value) {
            if (substr_count($key, $label) > 0) {
                return $value;
            }
        }
        return false;
    }

    /**
     * Skip add customily properties to cart item
     * @param $product_id
     * @param $properties
     * @param $cart_item_data
     */
    private function skip_add_properties_to_item($product_id, $properties, $cart_item_data) {
        $local_product_id = $this->find_property('_customily-woocommerce-product-id', $properties);
        if (
            !empty($local_product_id) &&
            (int)$local_product_id == (int)$product_id &&
            !isset($cart_item_data['bundled_item_id'])
        ) {
            $this->can_load_customily = $this->can_load_customily ?? $this->can_load_customily($product_id);
            return !$this->can_load_customily;
        }
        return true;
    }

    /**
     * @param $key
     * @return false|mixed
     */
    private function get_query_param($key) {
        return !empty($_GET[$key]) ? sanitize_text_field(wp_unslash($_GET[$key])) : false;
    }

    /**
     * @param array $properties
     * @return false|mixed|string
     */
    private function use_app_version2($properties = []) {
        if (count($properties) > 0) {
            $property_key = 'property___';
            $filtered_array = array_filter(array_keys($properties), function ($key) use ($property_key) {
                return substr_count($key, $property_key) > 0;
            });
            return count($filtered_array) > 0 ? '1' : false;
        }

        return $this->get_query_param('useAppV2');
    }

    /**
     * Get customily preview property html
     * @param $property
     * @param $return_default_label
     * @return array|string[]
     */
    private function get_preview_link_property($property, $return_default_label = false) {
        $label = $return_default_label
            ? '_customily-preview'
            : $this->get_setting_value('preview_link_label', $this->app_settings);
        $link = $property['value'];
        $config = get_option('cl_preview_link_html');
        $config = json_decode($config);

        if ($config && !empty($config->src)) {
            $result = ['label' => '', 'value' => ''];
            $width = $config->width ?? 30;
            $height = $config->height ?? 30;
            $container_styles = $config->container_styles ?? '';
            $link_styles = $config->link_styles ?? '';
            $icon_styles = $config->icon_styles ?? '';
            $result['label'] = $config->label ?? '';

            $image = "<img src=\"{$config->src}\" width=\"{$width}\" height=\"{$height}\" style=\"{$icon_styles}\; position: relative;\">";
            $value = '<a href="' . $link . '" target="_blank" style="' . $link_styles . '">' . $image . '</a>';

            if ($config->icon_first == '1') {
                $label = '<a href="' . $link . '" target="_blank" style="' . $link_styles . '">' . $label . '</a>';
                $result['value'] = "{$value}&nbsp;&nbsp;{$label}";
            } else {
                $result['value'] = "{$label}&nbsp;&nbsp;{$value}";
            }

            $result['label'] = '<div class"customily-preview-link" style="' . $container_styles . '">' . $result['label'] . '</div>';
            return $result;
        }

        return [
            'label' => $label,
            'value' => '<a href="' . $link . '" target="_blank">' . $this->get_setting_value('preview_link_text', $this->app_settings) . '</a>'
        ];
    }

    /**
     * Get customization url property html
     * @param $properties
     * @return false|mixed|string
     */
    private function get_customization_url_property($properties) {
        if (isset($properties['property[_customily-customization-url]'])) {
            $prop = $properties['property[_customily-customization-url]'];
        } else if (isset($properties['property___customily-customization-url'])) {
            $prop = $properties['property___customily-customization-url'];
        }

        if (!empty($prop)) {
            $show_customization_url_as_text = get_option('cl_show_customization_url_as_text');
            return $show_customization_url_as_text != 'y'
                ? '<a href="' . $prop['value'] . '" target="_blank">Link</a>'
                : $prop['value'];
        }

        return false;
    }
    /**
     * Get multi-file properties
     * @param $file_number
     * @return array
     */
    private function get_file_properties($file_number) {
        return array(
            "property___customily-production-url-{$file_number}",
            "property___customily-personalization-id-{$file_number}",
            "property___customily-eps-name-{$file_number}",
            "property___customily-eps-placeholder-{$file_number}"
        );
    }

    /**
     * Get multi-files properties
     * @param $file_numbers
     * @return array
     */
    private function get_multi_files_properties($file_numbers) {
        $file_numbers_array = explode(',', $file_numbers);
        $result = [];
        foreach ($file_numbers_array as $file_number) {
            $file_number = (int) $file_number;
            if ($file_number > 1) {
                $file_properties = $this->get_file_properties($file_number);
                $result = array_merge($result, $file_properties);
            }
        }
        return $result;
    }
}